//! INes 2.0 Format implementation

use crate::rom::{CpuTiming, NametableMirroring, Rom, RomLoader, RomReadError};

pub struct INesHeader {
    prg_rom_size_lsb: u8,
    chr_rom_size_lsb: u8,

    // Flags 6
    // D~7654 3210
    //   ---------
    //   NNNN FTBM
    //   |||| |||+-- Hard-wired nametable layout
    //   |||| |||     0: Vertical arrangement ("mirrored horizontally") or mapper-controlled
    //   |||| |||     1: Horizontal arrangement ("mirrored vertically")
    //   |||| ||+--- "Battery" and other non-volatile memory
    //   |||| ||      0: Not present
    //   |||| ||      1: Present
    //   |||| |+--- 512-byte Trainer
    //   |||| |      0: Not present
    //   |||| |      1: Present between Header and PRG-ROM data
    //   |||| +---- Alternative nametables
    //   ||||        0: No
    //   ||||        1: Yes
    //   ++++------ Mapper Number D3..D0
    flag_6: u8,

    // Flags 7
    // D~7654 3210
    //   ---------
    //   NNNN 10TT
    //   |||| ||++- Console type
    //   |||| ||     0: Nintendo Entertainment System/Family Computer
    //   |||| ||     1: Nintendo Vs. System
    //   |||| ||     2: Nintendo Playchoice 10
    //   |||| ||     3: Extended Console Type
    //   |||| ++--- NES 2.0 identifier
    //   ++++------ Mapper Number D7..D4
    flag_7: u8,

    // Mapper MSB/Submapper
    // D~7654 3210
    //   ---------
    //   SSSS NNNN
    //   |||| ++++- Mapper number D11..D8
    //   ++++------ Submapper number
    mapper_msb_submapper: u8,

    // PRG-ROM/CHR-ROM size MSB
    // D~7654 3210
    //   ---------
    //   CCCC PPPP
    //   |||| ++++- PRG-ROM size MSB
    //   ++++------ CHR-ROM size MSB
    rom_sizes_msb: u8,

    // PRG-RAM/EEPROM size
    //  D~7654 3210
    //    ---------
    //    pppp PPPP
    //    |||| ++++- PRG-RAM (volatile) shift count
    //    ++++------ PRG-NVRAM/EEPROM (non-volatile) shift count
    //  If the shift count is zero, there is no PRG-(NV)RAM.
    //  If the shift count is non-zero, the actual size is
    //  "64 << shift count" bytes, i.e. 8192 bytes for a shift count of 7.
    prg_ram_eeprom_size: u8,

    // CHR-RAM size
    //  D~7654 3210
    //    ---------
    //    cccc CCCC
    //    |||| ++++- CHR-RAM size (volatile) shift count
    //    ++++------ CHR-NVRAM size (non-volatile) shift count
    //  If the shift count is zero, there is no CHR-(NV)RAM.
    //  If the shift count is non-zero, the actual size is
    //  "64 << shift count" bytes, i.e. 8192 bytes for a shift count of 7.
    chr_ram_size: u8,

    // CPU/PPU Timing
    // D~7654 3210
    //   ---------
    //   .... ..VV
    //          ++- CPU/PPU timing mode
    //               0: RP2C02 ("NTSC NES")
    //               1: RP2C07 ("Licensed PAL NES")
    //               2: Multiple-region
    //               3: UA6538 ("Dendy")
    cpu_ppu_timing: u8,

    // When Byte 7 AND 3 =1: Vs. System Type
    // D~7654 3210
    //   ---------
    //   MMMM PPPP
    //   |||| ++++- Vs. PPU Type
    //   ++++------ Vs. Hardware Type
    //
    // When Byte 7 AND 3 =3: Extended Console Type
    // D~7654 3210
    //   ---------
    //   .... CCCC
    //        ++++- Extended Console Type
    system_type: u8,

    // Miscellaneous ROMs
    // D~7654 3210
    //   ---------
    //   .... ..RR
    //          ++- Number of miscellaneous ROMs present
    misc_roms: u8,

    // Default Expansion Device
    // D~7654 3210
    //   ---------
    //   ..DD DDDD
    //     ++-++++- Default Expansion Device
    default_expansion_device: u8,
}

impl INesHeader {
    /// Identification characters at the start of a iNes2.0 ROM file - NES<EOF>
    const IDENTIFICATION: [u8; 4] = [0x4E, 0x45, 0x53, 0x1A];

    fn get_prg_ram_size(&self) -> (usize, usize) {
        let prg_ram_size = (self.prg_ram_eeprom_size & 0xF) as usize;
        let prg_nvram_size = (self.prg_ram_eeprom_size >> 4) as usize;

        (prg_ram_size, prg_nvram_size)
    }

    fn get_prg_rom_size(&self) -> usize {
        let prg_rom_size_lsb = self.prg_rom_size_lsb as usize;
        let prg_rom_size_msb = (self.rom_sizes_msb & 0x0F) as usize;
        if prg_rom_size_msb < 0xF {
            // LSB and MSB are the size of the PRG ROM in 16KiB units
            ((prg_rom_size_msb << 8) | prg_rom_size_lsb) << 14
        } else {
            // (2 ^ exponent) * multiplier
            let multiplier = (prg_rom_size_lsb & 0x2) * 2 + 1;
            let exponent = prg_rom_size_lsb >> 2;
            (1 << exponent) * multiplier
        }
    }

    fn get_chr_ram_size(&self) -> (usize, usize) {
        let chr_ram_shifts = (self.chr_ram_size & 0xF) as usize;
        let chr_nvram_shifts = (self.chr_ram_size >> 4) as usize;

        let chr_ram_size = match chr_ram_shifts {
            0 => 0,
            shifts => 64 << shifts,
        };

        let chr_nvram_size = match chr_nvram_shifts {
            0 => 0,
            shifts => 64 << shifts,
        };

        (chr_ram_size, chr_nvram_size)
    }

    fn get_chr_rom_size(&self) -> usize {
        let chr_rom_size_lsb = self.chr_rom_size_lsb as usize;
        let chr_rom_size_msb = (self.rom_sizes_msb >> 4) as usize;
        if chr_rom_size_msb < 0xF {
            // LSB and MSB are the size of the CHR ROM in 8KiB units
            ((chr_rom_size_msb << 8) | chr_rom_size_lsb) << 13
        } else {
            // (2 ^ exponent) * multiplier
            let multiplier = (chr_rom_size_lsb & 0x2) * 2 + 1;
            let exponent = chr_rom_size_lsb >> 2;
            (1 << exponent) * multiplier
        }
    }

    fn get_trainer_size(&self) -> usize {
        if self.flag_6 & 4 == 0 {
            0
        } else {
            512
        }
    }

    fn get_nametable_mirroring(&self) -> NametableMirroring {
        match self.flag_6 & 1 {
            0 => NametableMirroring::Horizontal,
            _ => NametableMirroring::Vertical,
        }
    }

    fn get_cpu_timing(&self) -> Result<CpuTiming, RomReadError> {
        Ok(match self.cpu_ppu_timing {
            0 => CpuTiming::Ntsc,
            1 => CpuTiming::Pal,
            2 => CpuTiming::Multiple,
            3 => CpuTiming::Dendy,
            _ => {
                return Err(RomReadError::InvalidHeader(format!(
                    "Invalid CPU timing mode: {}",
                    self.cpu_ppu_timing
                )))
            }
        })
    }

    fn get_mapper(&self) -> u16 {
        let mapper_low = (self.flag_6 >> 4) as u16;
        let mapper_mid = (self.flag_7 & 0xF0) as u16;
        let mapper_high = (self.mapper_msb_submapper & 0x0F) as u16;

        mapper_high << 8 | mapper_mid | mapper_low
    }

    fn get_sub_mapper(&self) -> u8 {
        self.mapper_msb_submapper >> 4
    }

    fn parse(header: &[u8]) -> Self {
        Self {
            prg_rom_size_lsb: header[4],
            chr_rom_size_lsb: header[5],
            flag_6: header[6],
            flag_7: header[7],
            mapper_msb_submapper: header[8],
            rom_sizes_msb: header[9],
            prg_ram_eeprom_size: header[10],
            chr_ram_size: header[11],
            cpu_ppu_timing: header[12],
            system_type: header[13],
            misc_roms: header[14],
            default_expansion_device: header[15],
        }
    }
}

impl RomLoader for INesHeader {
    fn format_matches(header: &[u8]) -> bool {
        for i in 0..4 {
            if header[i] != Self::IDENTIFICATION[i] {
                // This is not a header from a iNes 2.0 ROM
                return false;
            }
        }

        // if header[7] & 0b00001100 != 0b00001000 {
        //     // This is a iNes 1.0 ROM
        // }

        true
    }

    fn load(header: &[u8], file_data: &[u8]) -> Result<Rom, RomReadError> {
        let ines_header = Self::parse(header);
        let prg_rom_size = ines_header.get_prg_rom_size();
        let chr_rom_size = ines_header.get_chr_rom_size();
        let trainer_size = ines_header.get_trainer_size();

        let rom_size = file_data.len();
        let required_size = prg_rom_size + chr_rom_size + trainer_size;
        if rom_size < required_size {
            return Err(RomReadError::RomLengthMismatch(required_size, rom_size));
        }

        let (prg_ram_size, prg_nvram_size) = ines_header.get_prg_ram_size();
        let (chr_ram_size, chr_nvram_size) = ines_header.get_chr_ram_size();
        let nametable_mirroring = ines_header.get_nametable_mirroring();
        let cpu_timing = ines_header.get_cpu_timing()?;
        let mapper = ines_header.get_mapper();
        let sub_mapper = ines_header.get_sub_mapper();

        let prg_rom = file_data[trainer_size..prg_rom_size]
            .to_vec()
            .into_boxed_slice();
        let chr_rom = file_data
            [(trainer_size + prg_rom_size)..(trainer_size + prg_rom_size + chr_rom_size)]
            .to_vec()
            .into_boxed_slice();

        Ok(Rom {
            prg_ram_size,
            prg_nvram_size,
            prg_rom_size,
            prg_rom,
            chr_ram_size,
            chr_nvram_size,
            chr_rom_size,
            chr_rom,
            nametable_mirroring,
            cpu_timing,
            mapper,
            sub_mapper,
        })
    }
}
