use simplelog::info;
use std::cell::RefCell;
use std::rc::Rc;
use crate::cpu::Cpu;
use crate::mappers::get_mapper;
use crate::memory::MemoryBus;
use crate::rom::{Rom, RomReadError};
use crate::Clock;
use crate::ppu::Ppu;

const CPU_CLOCK_DIVISOR: usize = 12;
const CPU_CYCLE_PER_REFRESH: usize = MASTER_CYCLE_PER_REFRESH / CPU_CLOCK_DIVISOR;
const MASTER_CLOCK: usize = 21477272; // NTSC NES Master Clock (~21.47 MHz)
const MASTER_CYCLE_PER_REFRESH: usize = MASTER_CLOCK / REFRESH_RATE;
const PPU_CLOCK_DIVISOR: usize = 4;
const PPU_CYCLE_PER_REFRESH: usize = CPU_CLOCK_DIVISOR / PPU_CLOCK_DIVISOR;
const REFRESH_RATE: usize = 60; // 60 Hz

pub struct System {
    cpu: Cpu,
    memory_bus: Rc<RefCell<MemoryBus>>,
    ppu: Rc<RefCell<Ppu>>
}

impl System {
    pub fn new() -> Self {
        let memory_bus = Rc::new(RefCell::new(MemoryBus::new()));
        let cpu = Cpu::new(Rc::clone(&memory_bus));
        let ppu = Rc::new(RefCell::new(Ppu::new(Rc::clone(&memory_bus))));
        
        let weak_ppu = Rc::downgrade(&ppu);
        memory_bus.borrow_mut().register_observer(weak_ppu);
        
        Self { cpu, memory_bus, ppu }
    }

    pub fn insert_rom(&mut self, path: &str) -> Result<(), RomReadError> {
        let rom = Rom::read(path)?;
        let mapper = get_mapper(rom);

        self.memory_bus.borrow_mut().register_mapper(mapper);
        self.cpu.initialize();

        info!("Successfully inserted ROM into system");
        Ok(())
    }
}

impl Clock for System {
    fn cycle(&mut self) {
        for _cpu_cycle in 0usize..CPU_CYCLE_PER_REFRESH {
            self.cpu.cycle();

            for _ppu_cycle in 0usize..PPU_CYCLE_PER_REFRESH {
                self.ppu.borrow_mut().cycle();
            }
        }
    }
}
